from airflow import DAG

from ditto.templates.emr_hdi_template import EmrHdiDagTransformerTemplate
from ditto.transformers.subdag import CheckClusterSubDagTransformer


class CheckClusterEmr2HdiDagTransformerTemplate(EmrHdiDagTransformerTemplate):
    """
    A template which uses the :class:`~ditto.transformers.subdag.CheckClusterSubDagTransformer`

    .. seealso::

        See `examples/example_emr_job_flow_dag_2.py` for an example.
        You can find more examples in the unit tests at `tests/test_dag_transformations.py`
    """
    def __init__(self, src_dag: DAG, *args, **kwargs):
        super().__init__(src_dag, subdag_transformers=[CheckClusterSubDagTransformer],
                         *args, **kwargs)



from ditto.transformers.emr.emr_add_steps_op_transformer import EmrAddStepsOperatorTransformer
from ditto.transformers.emr.emr_create_job_flow_op_transformer import EmrCreateJobFlowOperatorTransformer
from ditto.transformers.emr.emr_job_flow_sensor_transformer import EmrJobFlowSensorTransformer
from ditto.transformers.emr.emr_step_sensor_transformer import EmrStepSensorTransformer
from ditto.transformers.emr.emr_terminate_job_flow_op_transformer import EmrTerminateJobFlowOperatorTransformer

__all__ = ["EmrAddStepsOperatorTransformer", "EmrCreateJobFlowOperatorTransformer",
           "EmrJobFlowSensorTransformer", "EmrStepSensorTransformer",
           "EmrTerminateJobFlowOperatorTransformer"]


from typing import List, Callable

import networkx as nx
from airflow import DAG
from airflow.contrib.operators.emr_create_job_flow_operator import EmrCreateJobFlowOperator
from airflow.operators.dummy_operator import DummyOperator

from airflow.models.taskinstance import TaskInstance
from ditto.api import SubDagTransformer, TaskMatcher, DAGFragment, TransformerDefaults
from ditto.matchers import PythonCallTaskMatcher, ClassTaskMatcher
from ditto.transformers.emr import EmrCreateJobFlowOperatorTransformer
from ditto.utils import TransformerUtils


class CheckClusterSubDagTransformer(SubDagTransformer):
    """
    Transforms a create-if-not-exists subdag pattern
    commonly used in EMR airflow DAGs to the corresponding
    pattern for an Azure HDI airflow DAG
    """
    def __init__(self, dag: DAG, defaults: TransformerDefaults):
        super().__init__(dag, defaults)
        #: the python callable which checks if the cluster exists.
        self.check_for_existing_emr_cluster = defaults.other_defaults['pycall_check_cluster']

    def get_sub_dag_matcher(self) -> List[TaskMatcher]:
        check_for_emr_cluster_op = PythonCallTaskMatcher(self.check_for_existing_emr_cluster)
        create_cluster_op = ClassTaskMatcher(EmrCreateJobFlowOperator)
        cluster_exists_op = ClassTaskMatcher(DummyOperator)
        get_cluster_id_op = PythonCallTaskMatcher(TaskInstance.xcom_pull)

        check_for_emr_cluster_op >> [create_cluster_op, cluster_exists_op]
        create_cluster_op >> get_cluster_id_op
        cluster_exists_op >> get_cluster_id_op

        return [check_for_emr_cluster_op]

    def transform(self, subdag: nx.DiGraph, parent_fragment: DAGFragment) -> DAGFragment:
        transformer = EmrCreateJobFlowOperatorTransformer(self.dag, self.defaults)
        return transformer.transform(
            TransformerUtils.find_matching_tasks(
                subdag, ClassTaskMatcher(EmrCreateJobFlowOperator))[0], parent_fragment)


from typing import List

from airflow import DAG
from airflow.contrib.sensors.emr_job_flow_sensor import EmrJobFlowSensor
from airflow.models import BaseOperator

from ditto.api import OperatorTransformer, TransformerDefaults, DAGFragment, UpstreamOperatorNotFoundException
from ditto.utils import TransformerUtils
from airflowhdi.operators import ConnectedAzureHDInsightCreateClusterOperator
from airflowhdi.sensors import AzureHDInsightClusterSensor


class EmrJobFlowSensorTransformer(OperatorTransformer[EmrJobFlowSensor]):
    """
    Transforms the sensor :class:`~airflow.contrib.operators.emr_job_flow_sensor.EmrJobFlowSensor`
    """
    def __init__(self, target_dag: DAG, defaults: TransformerDefaults):
        super().__init__(target_dag, defaults)

    def transform(self, src_operator: BaseOperator, parent_fragment: DAGFragment, upstream_fragments: List[DAGFragment]) -> DAGFragment:
        """
        This transformer assumes and relies on the fact that an upstream transformation
        of a :class:`~airflow.contrib.operators.emr_create_job_flow_operator.EmrCreateJobFlowOperator`
        has already taken place, since it needs to find the output of that transformation
        to get the `cluster_name` and `azure_conn_id` from that operator (which should have been a
        :class:`~airflowhdi.operators.AzureHDInsightCreateClusterOperator`)

        Creates a :class:`~airflowhdi.sensors.AzureHDInsightClusterSensor` in non-provisioning mode
        to monitor the cluster till it reaches a terminal state (cluster shutdown by user or failed).

        .. warning::

            We do not have a way to tell the HDInsight cluster to halt if a job has failed, unlike EMR.
            So the cluster will continue to run even on job failure. You have to add a terminate cluster
            operator on step failure through ditto itself.
        """
        create_op_task_id = TransformerUtils.get_task_id_from_xcom_pull(src_operator.job_flow_id)
        create_op: BaseOperator = \
            TransformerUtils.find_op_in_fragment_list(
                upstream_fragments,
                operator_type=ConnectedAzureHDInsightCreateClusterOperator,
                task_id=create_op_task_id)

        if not create_op:
            raise UpstreamOperatorNotFoundException(ConnectedAzureHDInsightCreateClusterOperator,
                                                    EmrJobFlowSensor)

        monitor_cluster_op = AzureHDInsightClusterSensor(create_op.cluster_name,
                                                         azure_conn_id=create_op.azure_conn_id,
                                                         poke_interval=5,
                                                         task_id=f"{create_op.task_id}_monitor_cluster",
                                                         dag=self.dag)

        self.copy_op_attrs(monitor_cluster_op, src_operator)
        self.sign_op(monitor_cluster_op)

        return DAGFragment([monitor_cluster_op])


import inspect
from typing import Type

from airflow.models import BaseOperator

from ditto.api import TaskMatcher


class ClassTaskMatcher(TaskMatcher):
    """
    Matches a given operator based on the class being the same
    """
    def __init__(self, operator_class: Type):
        """
        :param operator_class: the class which should be checked for equality
        """
        super().__init__()
        self.operator_class = operator_class

    def does_match(self, task: BaseOperator) -> bool:
        """
        Matches this operator's entire ancestry of classes
        to the `operator_class` to match against

        :param task:
        :return:
        """
        for task_cls in inspect.getmro(task.__class__):
            if task_cls is self.operator_class:
                return True


from typing import List

from airflow import DAG
from airflow.models import BaseOperator
from airflow.operators.sensors import S3KeySensor

from ditto.api import OperatorTransformer, TransformerDefaults, DAGFragment, TransformerException
from airflowhdi.sensors import AzureDataLakeStorageGen1WebHdfsSensor


class S3KeySensorAdlsGen1OperatorTransformer(OperatorTransformer[S3KeySensor]):
    """
    Transforms a :class:`~airflow.operators.sensors.S3KeySensor`
    into a :class:`~airflowhdi.sensors.AzureDataLakeStorageGen1WebHdfsSensor`
    """
    def __init__(self, target_dag: DAG, defaults: TransformerDefaults):
        super().__init__(target_dag, defaults)

    def transform(self, src_operator: BaseOperator, parent_fragment: DAGFragment, upstream_fragments: List[DAGFragment]) -> DAGFragment:
        """
        You need to add the ``adls_conn_id`` to the source operator (or preferably DAG) for this to work.
        The ``glob_path`` for the ADLS sensor is coped from the ``bucket_key`` of the s3 sensor,
        so make sure that is templatized for changing between `s3://` and `adls://` paths using config
        """
        s3_key_sensor: S3KeySensor = src_operator
        adls_conn_id = s3_key_sensor.params.get('adls_conn_id', None)
        if not adls_conn_id:
            adls_conn_id = self.dag.params.get('adls_conn_id', None)

        if not adls_conn_id:
            raise TransformerException("Could not find adls_conn_id in operator or DAG params")

        adls_gen1_sensor_op = AzureDataLakeStorageGen1WebHdfsSensor(
            task_id=src_operator.task_id,
            azure_data_lake_conn_id=adls_conn_id,
            glob_path=s3_key_sensor.bucket_key,
            dag=self.dag
        )

        self.copy_op_attrs(adls_gen1_sensor_op, src_operator)
        self.sign_op(adls_gen1_sensor_op)

        return DAGFragment([adls_gen1_sensor_op])


import inspect
from typing import Dict, Callable, Type

from airflow.models import BaseOperator
from airflow.operators.python_operator import PythonOperator, BranchPythonOperator

from ditto.api import TransformerResolver, OperatorTransformer


class PythonCallTransformerResolver(TransformerResolver):
    """
    Similar to the :class:`~ditto.matchers.PythonCallTaskMatcher`, but finds a transformer
    for a operator using the same matching pattern.
    """
    def __init__(self, callable_transformers: Dict[Callable, Type[OperatorTransformer]],
                 nested_search: bool = True):
        """
        :param callable_transformers: a map of python callables and matching transformers
        :param nested_search: whether to search the method source code itself for the callable
        """
        self.transformers_for_callables = callable_transformers
        self.nested_search = nested_search

    def resolve_transformer(self, task: BaseOperator) -> Type[OperatorTransformer]:
        """
        see the behavior of :meth:`~ditto.matchers.PythonCallTaskMatcher.does_match`

        :param task: the task to resolve the transformer for
        :return: the resolved transformer
        """
        if isinstance(task, PythonOperator) or isinstance(task, BranchPythonOperator):
            python_op: PythonOperator = task
            py_callable = python_op.python_callable
            src_lines = inspect.getsourcelines(py_callable)[0]

            if self.nested_search:
                for line in src_lines:
                    for find_callable in self.transformers_for_callables:
                        if find_callable.__name__ in line:
                            return self.transformers_for_callables[find_callable]
            else:
                for find_callable in self.transformers_for_callables:
                    if find_callable is py_callable:
                        return self.transformers_for_callables[find_callable]


from ditto.templates.emr_hdi_template import EmrHdiDagTransformerTemplate
from ditto.templates.emr_hdi_check_cluster_template import CheckClusterEmr2HdiDagTransformerTemplate

__all__ = ["EmrHdiDagTransformerTemplate", "CheckClusterEmr2HdiDagTransformerTemplate"]


from typing import List

from airflow import DAG
from airflow.contrib.sensors.wasb_sensor import WasbPrefixSensor
from airflow.models import BaseOperator
from airflow.operators.sensors import S3KeySensor

from ditto.api import OperatorTransformer, TransformerDefaults, DAGFragment, TransformerException
from airflowhdi.sensors import WasbWildcardPrefixSensor


class S3KeySensorBlobOperatorTransformer(OperatorTransformer[S3KeySensor]):
    """
    Transforms a :class:`~airflow.operators.sensors.S3KeySensor`
    into either a :class:`~airflowhdi.sensors.WasbWildcardPrefixSensor`
    or a :class:`~airflow.contrib.sensors.wasb_sensor.WasbPrefixSensor`
    """
    def __init__(self, target_dag: DAG, defaults: TransformerDefaults):
        super().__init__(target_dag, defaults)

    def transform(self, src_operator: BaseOperator, parent_fragment: DAGFragment, upstream_fragments: List[DAGFragment]) -> DAGFragment:
        """
        You need to add the ``wasb_conn_id`` to the source operator (or preferably DAG) for this to work.
        The ``container_name`` and ``prefix`` for the blob based sensors are coped from the ``bucket_name``
        and ``bucket_key`` of the s3 sensor, so make sure they are templatized for changing between
        `s3://` and `wasb://` paths, etc. using config
        """
        s3_key_sensor: S3KeySensor = src_operator
        wasb_conn_id = s3_key_sensor.params.get('wasb_conn_id', None)
        if not wasb_conn_id:
            wasb_conn_id = self.dag.params.get('wasb_conn_id', None)

        if not wasb_conn_id:
            raise TransformerException("Could not find wasb_conn_id in operator or DAG params")

        if s3_key_sensor.wildcard_match:
            wasb_sensor_op = WasbWildcardPrefixSensor(
                task_id=src_operator.task_id,
                wasb_conn_id=wasb_conn_id,
                container_name=s3_key_sensor.bucket_name,
                wildcard_prefix=s3_key_sensor.bucket_key,
                dag=self.dag
            )
        else:
            wasb_sensor_op = WasbPrefixSensor(
                task_id=src_operator.task_id,
                wasb_conn_id=wasb_conn_id,
                container_name=s3_key_sensor.bucket_name,
                prefix=s3_key_sensor.bucket_key,
                dag=self.dag
            )

        self.copy_op_attrs(wasb_sensor_op, src_operator)
        self.sign_op(wasb_sensor_op)

        return DAGFragment([wasb_sensor_op])


from typing import List

from airflow import DAG
from airflow.contrib.operators.emr_add_steps_operator import EmrAddStepsOperator
from airflow.models import BaseOperator
from airflow.operators.dummy_operator import DummyOperator
from airflow.utils.trigger_rule import TriggerRule

from ditto.api import OperatorTransformer, TransformerDefaults, DAGFragment, UpstreamOperatorNotFoundException
from ditto.utils import TransformerUtils
from airflowhdi.operators.azure_hdinsight_create_cluster_operator import ConnectedAzureHDInsightCreateClusterOperator
from airflowhdi.operators.azure_hdinsight_ssh_operator import AzureHDInsightSshOperator
from airflowhdi.operators import LivyBatchOperator


class EmrAddStepsOperatorTransformer(OperatorTransformer[EmrAddStepsOperator]):
    """
    Transforms the operator :class:`~airflow.contrib.operators.emr_add_steps_operator.EmrAddStepsOperator`
    """

    #: default livy proxy user
    DEFAULT_PROXY_USER = "admin"

    #: default spark configuration to use
    DEFAULT_SPARK_CONF = {'spark.shuffle.compress': 'false'}

    #: number of mappers to use in case its a distcp
    HADOOP_DISTCP_DEFAULT_MAPPERS = 100

    def __init__(self, target_dag: DAG, defaults: TransformerDefaults):
        super().__init__(target_dag, defaults)
        self.proxy_user = self.defaults.other_defaults.get('proxy_user', EmrAddStepsOperatorTransformer.DEFAULT_PROXY_USER)
        self.spark_conf = self.defaults.other_defaults.get('spark_conf', EmrAddStepsOperatorTransformer.DEFAULT_SPARK_CONF)

    @staticmethod
    def get_target_step_task_id(add_step_task_id, add_step_step_id):
        """
        generates a task_id for the transformed output operators
        for each input EMR step
        """
        return f"{add_step_task_id}_{add_step_step_id}"

    def transform(self, src_operator: BaseOperator, parent_fragment: DAGFragment, upstream_fragments: List[DAGFragment]) -> DAGFragment:
        """
        This transformer assumes and relies on the fact that an upstream transformation
        of a :class:`~airflow.contrib.operators.emr_create_job_flow_operator.EmrCreateJobFlowOperator`
        has already taken place, since it needs to find the output of that transformation
        to get the `cluster_name` and `azure_conn_id` from that operator (which should have been a
        :class:`~airflowhdi.operators.AzureHDInsightCreateClusterOperator`)

        It then goes through the EMR steps of this :class:`~airflow.contrib.operators.emr_add_steps_operator.EmrAddStepsOperator`
        and creates a :class:`~airflowhdi.operators.LivyBatchOperator` or an :class:`~airflowhdi.operators.AzureHDInsightSshOperator`
        for each corresponding step, based on grokking the step's params and figuring out whether its a spark job being run
        on an arbitrary hadoop command like `distcp`, `hdfs` or the like.

        .. note::

            This transformer creates multiple operators from a single source operator

        .. note::

            The spark configuration for the livy spark job are derived from `step['HadoopJarStep']['Properties']` of the EMR step,
            or could even be specified at the cluster level itself when transforming the job flow
        """
        create_op_task_id = TransformerUtils.get_task_id_from_xcom_pull(src_operator.job_flow_id)
        create_op: BaseOperator = \
            TransformerUtils.find_op_in_fragment_list(
                upstream_fragments,
                operator_type=ConnectedAzureHDInsightCreateClusterOperator,
                task_id=create_op_task_id)

        if not create_op:
            raise UpstreamOperatorNotFoundException(ConnectedAzureHDInsightCreateClusterOperator,
                                                    EmrAddStepsOperator)

        emr_add_steps_op: EmrAddStepsOperator = src_operator
        dag_fragment_steps = []
        steps_added_op = DummyOperator(
            task_id=f"{emr_add_steps_op.task_id}_added",
            dag=self.dag)
        self.sign_op(steps_added_op)

        for step in emr_add_steps_op.steps:
            name = step['Name']

            ssh_command = None
            livy_file = None
            livy_arguments = None
            livy_main_class = None

            if 'command-runner' in step['HadoopJarStep']['Jar']:
                command_runner_cmd = step['HadoopJarStep']['Args']
                if '/usr/bin/spark-submit' in command_runner_cmd[0]:
                    livy_file = command_runner_cmd[1]
                    livy_arguments = command_runner_cmd[2:]
                elif 's3-dist-cp' in command_runner_cmd[0]:
                    src = None
                    dest = None
                    for arg in command_runner_cmd[1:]:
                        if arg.startswith('--src='):
                            src = arg.split("--src=", 1)[1]
                        if arg.startswith('--dest='):
                            dest = arg.split("--dest=", 1)[1]
                    mappers = EmrAddStepsOperatorTransformer.HADOOP_DISTCP_DEFAULT_MAPPERS
                    ssh_command = f"hadoop distcp -m {mappers} {src} {dest}"
                elif 'hdfs' in command_runner_cmd[0]:
                    ssh_command = " ".join(command_runner_cmd)
                else:
                    raise Exception("This kind of step is not supported right now", command_runner_cmd[0])
            else:
                livy_file = step['HadoopJarStep']['Jar']
                livy_arguments = step['HadoopJarStep']['Args']
                livy_main_class = step['HadoopJarStep'].get('MainClass', None)

            if 'Properties' in step['HadoopJarStep']:
                properties = ""
                for key, val in step['HadoopJarStep']['Properties']:
                    properties += f"-D{key}={val} "

                self.spark_conf['spark.executor.extraJavaOptions'] = properties
                self.spark_conf['spark.driver.extraJavaOptions'] = properties

            target_step_task_id = EmrAddStepsOperatorTransformer.get_target_step_task_id(emr_add_steps_op.task_id, emr_add_steps_op.steps.index(step))

            if ssh_command is not None:
                step_op = AzureHDInsightSshOperator(
                    cluster_name=create_op.cluster_name,
                    azure_conn_id=create_op.azure_conn_id,
                    command=ssh_command,
                    task_id=target_step_task_id,
                    dag=self.dag
                )
            else:
                step_op = LivyBatchOperator(
                    name=name,
                    file=livy_file,
                    arguments=livy_arguments,
                    class_name=livy_main_class,
                    azure_conn_id=create_op.azure_conn_id,
                    cluster_name=create_op.cluster_name,
                    proxy_user=self.proxy_user,
                    conf=self.spark_conf,
                    task_id=target_step_task_id,
                    dag=self.dag
                )
            self.copy_op_attrs(step_op, emr_add_steps_op)
            self.sign_op(step_op)
            step_op.trigger_rule = TriggerRule.ALL_SUCCESS

            step_op.set_downstream(steps_added_op)
            dag_fragment_steps.append(step_op)

        return DAGFragment(dag_fragment_steps)


import copy
from typing import List

from airflow import DAG
from airflow.models import BaseOperator

from ditto.api import OperatorTransformer, TransformerDefaults, DAGFragment


class CopyTransformer(OperatorTransformer):
    """
    Deep copies the source operator, clears it of its connections
    to the source DAG, and any upstream/downstream tasks and assigns
    it to the `target_dag`.

    .. note::
        If you face issues with serialization of the transformed operator
        by airflow, you can exclude some attributes to be deep copied
        by overriding the following attribute of the source operator:

        >>> input_operator.shallow_copy_attrs += ('_dag',)
    """
    def __init__(self, target_dag: DAG, defaults: TransformerDefaults):
        super().__init__(target_dag, defaults)

    def transform(self, input_operator: BaseOperator, parent_fragment: DAGFragment, upstream_fragments: List[DAGFragment]) -> DAGFragment:
        copied_op = copy.deepcopy(input_operator)
        copied_op._dag = None
        copied_op._upstream_task_ids.clear()
        copied_op._downstream_task_ids.clear()
        copied_op.dag = self.dag
        return DAGFragment([copied_op])


from ditto.resolvers.class_transformer_resolver import ClassTransformerResolver
from ditto.resolvers.ancestral_class_transformer_resolver import AncestralClassTransformerResolver
from ditto.resolvers.python_call_transformer_resolver import PythonCallTransformerResolver

__all__ = ["ClassTransformerResolver", "AncestralClassTransformerResolver",
           "PythonCallTransformerResolver"]


from typing import Dict, Type, List

from airflow import DAG
from airflow.contrib.operators.emr_add_steps_operator import EmrAddStepsOperator
from airflow.contrib.operators.emr_create_job_flow_operator import EmrCreateJobFlowOperator
from airflow.contrib.operators.emr_terminate_job_flow_operator import EmrTerminateJobFlowOperator
from airflow.contrib.sensors.emr_job_flow_sensor import EmrJobFlowSensor
from airflow.contrib.sensors.emr_step_sensor import EmrStepSensor
from airflow.models import BaseOperator
from airflow.operators.sensors import S3KeySensor

from ditto.api import TransformerDefaultsConf, OperatorTransformer, TransformerResolver, SubDagTransformer
from ditto.ditto import AirflowDagTransformer
from ditto.transformers.emr import EmrCreateJobFlowOperatorTransformer
from ditto.transformers.emr import EmrJobFlowSensorTransformer
from ditto.transformers.emr import EmrAddStepsOperatorTransformer
from ditto.transformers.emr import EmrStepSensorTransformer
from ditto.transformers.emr import EmrTerminateJobFlowOperatorTransformer
from ditto.transformers.s3 import S3KeySensorBlobOperatorTransformer
from ditto.resolvers import ClassTransformerResolver, AncestralClassTransformerResolver


class EmrHdiDagTransformerTemplate(AirflowDagTransformer):
    """
    This is the defacto template to use for converting an EMR-based airflow DAG
    to an Azure HDInsight based airlow DAG in ditto, unless you encounter
    more complex patterns, in which case you can always create your own template.

    This is easily sub-classable.

    .. seealso::

        See `examples/example_emr_job_flow_dag.py` for an example.
        You can find more examples in the unit tests at `tests/test_dag_transformations.py`
    """
    def __init__(self, target_dag: DAG,
                 transformer_defaults: TransformerDefaultsConf = None,
                 operator_transformers: Dict[Type[BaseOperator], Type[OperatorTransformer]] = None,
                 transformer_resolvers: List[TransformerResolver] = None,
                 subdag_transformers: List[Type[SubDagTransformer]] = None,
                 **kwargs):
        _operator_transformers = {
            EmrCreateJobFlowOperator: EmrCreateJobFlowOperatorTransformer,
            EmrJobFlowSensor: EmrJobFlowSensorTransformer,
            EmrAddStepsOperator: EmrAddStepsOperatorTransformer,
            EmrStepSensor: EmrStepSensorTransformer,
            EmrTerminateJobFlowOperator: EmrTerminateJobFlowOperatorTransformer,
            S3KeySensor: S3KeySensorBlobOperatorTransformer
        }
        if operator_transformers:
            _operator_transformers.update(operator_transformers)

        _transformer_resolvers = [
            ClassTransformerResolver(_operator_transformers),
            AncestralClassTransformerResolver(_operator_transformers)]

        if transformer_resolvers:
            _transformer_resolvers.append(transformer_resolvers)

        super().__init__(target_dag,
                         transformer_defaults=transformer_defaults,
                         transformer_resolvers=_transformer_resolvers,
                         subdag_transformers=subdag_transformers,
                         **kwargs)



from typing import List

from airflow import DAG, settings
from airflow.contrib.operators.emr_create_job_flow_operator import EmrCreateJobFlowOperator
from airflow.models import Connection, BaseOperator

from ditto.api import OperatorTransformer, TransformerDefaults, DAGFragment
from airflowhdi.operators import AzureHDInsightCreateClusterOperator
from airflowhdi.sensors.azure_hdinsight_cluster_sensor import AzureHDInsightClusterSensor


class EmrCreateJobFlowOperatorTransformer(OperatorTransformer[EmrCreateJobFlowOperator]):
    """
    Transforms the operator :class:`~airflow.contrib.operators.emr_create_job_flow_operator.EmrCreateJobFlowOperator`
    """

    def __init__(self, target_dag: DAG, defaults: TransformerDefaults):
        super().__init__(target_dag, defaults)

    def get_cluster_name(self, src_operator: EmrCreateJobFlowOperator):
        """
        get the cluster name from the EMR job flow and it's connection
        """
        emr_conn_id = src_operator.emr_conn_id
        session = settings.Session()
        emr_conn = session.query(Connection).filter(Connection.conn_id == emr_conn_id).first()
        config = emr_conn.extra_dejson.copy()
        config.update(src_operator.job_flow_overrides)
        return config['Name']

    def transform(self, src_operator: BaseOperator, parent_fragment: DAGFragment, upstream_fragments: List[DAGFragment] = None) -> DAGFragment:
        """
        Copies the :class:`~airflowhdi.operators.AzureHDInsightCreateClusterOperator` from the given
        :paramref:`~ditto.api.TransformerDefaults.default_operator` provided and attaches it to the
        target DAG after transferring airflow attributes from the source
        :class:`~airflow.contrib.operators.emr_create_job_flow_operator.EmrCreateJobFlowOperator`

        Relies on the default op as it is not possible to translate an EMR cluster spec to an HDInsight
        cluster spec automatically, so it is best to accept that operator from the user itself

        What it does do though is attach a :class:`~airflowhdi.sensors.AzureHDInsightClusterSensor`
        to monitor the provisioning of this newly created cluster.

        Since the HDInsight management client is idempotent, it does not matter if the cluster already exists
        and the operator simply moves on if that is the case.
        """
        create_cluster_op: AzureHDInsightCreateClusterOperator = self.get_default_op(src_operator)
        if create_cluster_op is None:
            raise Exception("This transformer needs a default output operator")
        create_cluster_op.dag = self.dag
        create_cluster_op.cluster_name = self.get_cluster_name(src_operator)
        self.sign_op(create_cluster_op)

        monitor_provisioning_op = AzureHDInsightClusterSensor(create_cluster_op.cluster_name,
                                                              azure_conn_id=create_cluster_op.azure_conn_id,
                                                              poke_interval=5,
                                                              provisioning_only=True,
                                                              task_id=f"{create_cluster_op.task_id}_monitor_provisioning",
                                                              dag=self.dag)
        self.copy_op_attrs(monitor_provisioning_op, src_operator)
        self.sign_op(monitor_provisioning_op)
        create_cluster_op.set_downstream(monitor_provisioning_op)
        return DAGFragment([create_cluster_op])


from ditto.transformers.s3.s3_key_sensor_blob_transformer import S3KeySensorBlobOperatorTransformer
from ditto.transformers.s3.s3_key_sensor_adlsgen1_transformer import S3KeySensorAdlsGen1OperatorTransformer

__all__ = ["S3KeySensorBlobOperatorTransformer", "S3KeySensorAdlsGen1OperatorTransformer"]


import inspect
from typing import Type

from airflow.models import BaseOperator

from ditto.api import OperatorTransformer
from ditto.resolvers import ClassTransformerResolver


class AncestralClassTransformerResolver(ClassTransformerResolver):
    """
    Extends the :class:`~ditto.resolvers.ClassTransformerResolver` by matching
    on all the ancestor classes of the given operator
    """
    def resolve_transformer(self, task: BaseOperator) -> Type[OperatorTransformer]:
        for task_cls in inspect.getmro(task.__class__):
            transformer_cl = self.get_transformer_for_class(task_cls)
            if transformer_cl:
                return transformer_cl


from typing import Dict, Type

from airflow.models import BaseOperator

from ditto.api import TransformerResolver, OperatorTransformer


class ClassTransformerResolver(TransformerResolver):
    """
    Finds a resolver for a transformer based on the class of the source operator given
    """
    def __init__(self, operator_transformers: Dict[Type[BaseOperator], Type[OperatorTransformer]]):
        """
        :param operator_transformers: a map of operator type and transformer type
        """
        super().__init__()
        self.operator_transformers = operator_transformers

    def get_transformer_for_class(self, task_cls: Type) -> Type[OperatorTransformer]:
        """
        return the transformer type for a given operator type

        :param task_cls: the operator type (class)
        :return: the transformer type
        """
        if task_cls in self.operator_transformers:
            transformer_cl = self.operator_transformers[task_cls]
            return transformer_cl

    def resolve_transformer(self, task: BaseOperator) -> Type[OperatorTransformer]:
        return self.get_transformer_for_class(task.__class__)


import inspect
import logging
from typing import Callable

from airflow.models import BaseOperator
from airflow.operators.python_operator import PythonOperator, BranchPythonOperator

from ditto.api import TaskMatcher

log = logging.getLogger(__name__)


class PythonCallTaskMatcher(TaskMatcher):
    """
    Matches a :class:`PythonOperator <airflow.operators.python_operator.PythonOperator>` or :class:`BranchPythonOperator <airflow.operators.python_operator.BranchPythonOperator>`
    if their `python_callable`\'s are making the matching method calls or are the same as the method calls to match against.
    """
    def __init__(self, find_callable: Callable, nested_search: bool = True):
        """
        :param find_callable: the python method call to search for
        :param nested_search: if True, will search the source code of the `python_callable`
            of the python airflow operators given, for the `find_callable`, otherwise will
            just match on the `python_callable` itself
        """
        super().__init__()
        self.find_callable = find_callable
        self.nested_search = nested_search

    def does_match(self, task: BaseOperator) -> bool:
        """
        Uses the python `inspect` module to get the source code for a given
        python method if `nested_search` is enabled, otherwise just matches on the
        `python_callable` of the :class:`~airflow.operators.python_operator.PythonOperator` or
        :class:`~airflow.operators.python_operator.BranchPythonOperator`

        :param task: the operator to match against
        :return: whether or not it matched
        """
        if isinstance(task, PythonOperator) or isinstance(task, BranchPythonOperator):
            python_op: PythonOperator = task
            py_callable = python_op.python_callable

            if self.nested_search:
                try:
                    src_lines = inspect.getsourcelines(py_callable)[0]
                    for line in src_lines:
                        if self.find_callable.__name__ in line:
                            return True
                except TypeError:
                    log.warn("Could not get the source code for %s", py_callable)
                    return False
            else:
                if self.find_callable is py_callable:
                    return True


from typing import List

from airflow import DAG
from airflow.contrib.sensors.emr_step_sensor import EmrStepSensor
from airflow.models import BaseOperator
from airflow.operators.dummy_operator import DummyOperator

from ditto.api import OperatorTransformer, TransformerDefaults, DAGFragment, UpstreamOperatorNotFoundException
from ditto.utils import TransformerUtils
from ditto.transformers.emr import EmrAddStepsOperatorTransformer
from airflowhdi.operators import ConnectedAzureHDInsightCreateClusterOperator
from airflowhdi.operators import LivyBatchOperator
from airflowhdi.sensors.livy_batch_sensor import LivyBatchSensor


class EmrStepSensorTransformer(OperatorTransformer[EmrStepSensor]):
    """
    Transforms the sensor :class:`~airflow.contrib.operators.emr_step_sensor.EmrStepSensor`
    """

    def __init__(self, target_dag: DAG, defaults: TransformerDefaults):
        super().__init__(target_dag, defaults)

    def transform(self, src_operator: BaseOperator, parent_fragment: DAGFragment, upstream_fragments: List[DAGFragment]) -> DAGFragment:
        """
        This transformer assumes and relies on the fact that an upstream transformation
        of a :class:`~airflow.contrib.operators.emr_create_job_flow_operator.EmrCreateJobFlowOperator`
        has already taken place, since it needs to find the output of that transformation
        to get the `cluster_name` and `azure_conn_id` from that operator (which should have been a
        :class:`~airflowhdi.operators.AzureHDInsightCreateClusterOperator`)

        This transformer also requires than there would already be transformations of
        :class:`~airflow.contrib.operators.emr_add_steps_operator.EmrAddStepsOperator` to
        :class:`~airflowhdi.operators.LivyBatchOperator` or :class:`~airflowhdi.operators.AzureHDInsightSshOperator`
        in the `upstream_fragments` which can then be monitored by the output tasks of
        this transformer. It needs to search for those ops upstream to find their task IDs

        Adds :class:`~airflowhdi.sensors.LivyBatchSensor` if it was a livy spark job.
        There's no sensor required for a transformed :class:`~airflowhdi.operators.AzureHDInsightSshOperator`
        as it is synchronous.
        """
        create_op_task_id = TransformerUtils.get_task_id_from_xcom_pull(src_operator.job_flow_id)
        create_op: BaseOperator = \
            TransformerUtils.find_op_in_fragment_list(
                upstream_fragments,
                operator_type=ConnectedAzureHDInsightCreateClusterOperator,
                task_id=create_op_task_id)

        if not create_op:
            raise UpstreamOperatorNotFoundException(ConnectedAzureHDInsightCreateClusterOperator,
                                                    EmrStepSensor)

        emr_step_sensor_op: EmrStepSensor = src_operator

        emr_add_step_task_id = TransformerUtils.get_task_id_from_xcom_pull(emr_step_sensor_op.step_id)
        emr_add_step_step_id = TransformerUtils.get_list_index_from_xcom_pull(emr_step_sensor_op.step_id)
        target_step_task_id = EmrAddStepsOperatorTransformer.get_target_step_task_id(emr_add_step_task_id, emr_add_step_step_id)

        add_step_op: BaseOperator = \
            TransformerUtils.find_op_in_fragment_list_strict(
                upstream_fragments,
                task_id=target_step_task_id)

        if isinstance(add_step_op, LivyBatchOperator):
            step_sensor_op = LivyBatchSensor(
                batch_id=f"{{{{ task_instance.xcom_pull('{target_step_task_id}', key='return_value') }}}}",
                task_id = emr_step_sensor_op.task_id,
                azure_conn_id=create_op.azure_conn_id,
                cluster_name=create_op.cluster_name,
                verify_in="yarn",
                dag=self.dag
            )
        else:
            # don't need a sensor for the ssh operator
            step_sensor_op = DummyOperator(task_id=emr_step_sensor_op.task_id, dag=self.dag)

        self.copy_op_attrs(step_sensor_op, emr_step_sensor_op)
        self.sign_op(step_sensor_op)

        return DAGFragment([step_sensor_op])


from setuptools import setup, find_namespace_packages

# read the contents of your README file
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='airflow-ditto',
    version='0.0.1.2',
    author="Angad Singh",
    author_email="angad.singh@trufactor.io",
    description="An airflow DAG transformation framework",
    long_description=long_description,
    long_description_content_type='text/markdown',
    license="Apache Software License (http://www.apache.org/licenses/LICENSE-2.0)",
    url="https://github.com/angadsingh/airflow-ditto",
    packages = find_namespace_packages(include=["ditto.*"]),
    install_requires=[
        'deepdiff==4.3.2',
        'networkx',
        'apache-airflow>=1.10.10,<=2.*',
        'matplotlib',
        'airflow-hdinsight',
        'boto3',
        'parameterized',
        'alchemy-mock'
    ],
    classifiers=[
        'Programming Language :: Python :: 3',
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7'
)


from typing import List

from airflow import DAG
from airflow.contrib.operators.emr_terminate_job_flow_operator import EmrTerminateJobFlowOperator
from airflow.models import BaseOperator
from airflow.utils.trigger_rule import TriggerRule

from ditto.api import OperatorTransformer, TransformerDefaults, DAGFragment, UpstreamOperatorNotFoundException
from ditto.utils import TransformerUtils
from airflowhdi.operators import ConnectedAzureHDInsightCreateClusterOperator
from airflowhdi.operators import AzureHDInsightDeleteClusterOperator


class EmrTerminateJobFlowOperatorTransformer(OperatorTransformer[EmrTerminateJobFlowOperator]):
    """
    Transforms the operator :class:`~airflow.contrib.operators.emr_terminate_job_flow_operator.EmrTerminateJobFlowOperator`
    """

    def __init__(self, target_dag: DAG, defaults: TransformerDefaults):
        super().__init__(target_dag, defaults)

    def transform(self, src_operator: BaseOperator, parent_fragment: DAGFragment, upstream_fragments: List[DAGFragment]) -> DAGFragment:
        """
        This transformer assumes and relies on the fact that an upstream transformation
        of a :class:`~airflow.contrib.operators.emr_create_job_flow_operator.EmrCreateJobFlowOperator`
        has already taken place, since it needs to find the output of that transformation
        to get the `cluster_name` and `azure_conn_id` from that operator (which should have been a
        :class:`~airflowhdi.operators.AzureHDInsightCreateClusterOperator`)

        Creates a :class:`~airflowhdi.operators.AzureHDInsightDeleteClusterOperator` to terminate
        the cluster
        """
        create_op_task_id = TransformerUtils.get_task_id_from_xcom_pull(src_operator.job_flow_id)
        create_op: BaseOperator = \
            TransformerUtils.find_op_in_fragment_list(
                upstream_fragments,
                operator_type=ConnectedAzureHDInsightCreateClusterOperator,
                task_id=create_op_task_id)

        if not create_op:
            raise UpstreamOperatorNotFoundException(ConnectedAzureHDInsightCreateClusterOperator,
                                                    EmrTerminateJobFlowOperator)

        emr_terminate_op: EmrTerminateJobFlowOperator = src_operator
        terminate_cluster_op = AzureHDInsightDeleteClusterOperator(task_id=emr_terminate_op.task_id,
                                                                   azure_conn_id=create_op.azure_conn_id,
                                                                   cluster_name=create_op.cluster_name,
                                                                   dag=self.dag)

        self.copy_op_attrs(terminate_cluster_op, src_operator)
        self.sign_op(terminate_cluster_op)
        terminate_cluster_op.trigger_rule = TriggerRule.ALL_DONE

        return DAGFragment([terminate_cluster_op])


from typing import List

from airflow.models import BaseOperator

from ditto.api import OperatorTransformer, DAGFragment


class IdentityTransformer(OperatorTransformer):
    """
    Re-uses the source operator as is as the transformed operator
    Just clears its connections to the source DAG and adds it
    to the target DAG

    .. warning::
        Use this transformer with caution. Reusing an operator
        can have weird side effects

    """
    def transform(self, input_operator: BaseOperator, parent_fragment: DAGFragment, upstream_fragments: List[DAGFragment]) -> DAGFragment:
        input_operator._dag = None
        input_operator._upstream_task_ids.clear()
        input_operator._downstream_task_ids.clear()
        input_operator.dag = self.dag
        return DAGFragment([input_operator])